@@ -1,28 +1,26 @@
 #!/usr/bin/env python3
 
 import os
-
-from flask import Flask, make_response, request, jsonify, abort
+import threading
+import logging
+import time
+import traceback
+import mimetypes
+from hashlib import sha256
 from pathlib import Path
 from functools import partial
 
+from flask import Flask, make_response, request, jsonify, abort
+from sqlalchemy import (MetaData, Table, Column, CHAR, VARCHAR, DateTime,
+                        LargeBinary, BOOLEAN)
 from pyaltt2.config import load_yaml, config_value
 from pyaltt2.db import Database
 from pyaltt2.res import ResourceStorage
 from pyaltt2.converters import val_to_boolean
-
-import pyaltt2.crypto as cr
-
-import threading
-import logging
-import time
-
 from datetime import datetime, timedelta
-
+import pyaltt2.crypto as cr
 import pytz
 
-import traceback
-
 logger = logging.getLogger('gunicorn.error')
 
 dir_me = Path(__file__).absolute().parents[1]
@@ -61,34 +59,44 @@ def ping():
 @app.route('/u', methods=['POST'])
 def upload():
     if request.headers.get('x-auth-key') != UPLOAD_KEY:
-        abort(403)
-    elif 'file' not in request.files:
-        abort(400)
-    f = request.files['file']
-    if not f.name:
-        abort(400)
+        return make_response('Invalid upload key', 403)
+    f = request.files.get('file', request.form.get('file'))
+    if f is None:
+        return make_response('File not uploaded', 403)
+    filename = request.form.get('fname', f.filename)
+    if filename is None:
+        return make_response('File name not specified', 403)
     esecs = int(request.form.get('expires', config['default-expires']))
     expires = (datetime.now() + timedelta(seconds=esecs)).replace(
         tzinfo=pytz.timezone(time.tzname[0]))
+    data = f.stream.read()
+    sha256sum_gen = sha256()
+    sha256sum_gen.update(data)
+    sha256sum = sha256sum_gen.hexdigest()
+    received_sha256sum = request.form.get('sha256sum')
+    if received_sha256sum and received_sha256sum != sha256sum:
+        return make_response('Checksum does not match', 422)
     file_id = cr.gen_random_str(16)
     file_key = cr.gen_random_str(16)
-    file_name = os.path.basename(f.filename)
+    filename = os.path.basename(filename)
     oneshot = val_to_boolean(request.form.get('oneshot', False))
     engine = cr.Rioja(file_key, bits=256)
-    contents = engine.encrypt(f.stream.read(), b64=False)
-    response = make_response('', 201)
-    response.headers['Location'] = (f'{EXTERNAL_URL}/d/{file_id}/'
-                                    f'{file_key}/{file_name}')
+    contents = engine.encrypt(data, b64=False)
+    location = (f'{EXTERNAL_URL}/d/{file_id}/' f'{file_key}/{filename}')
+    response = make_response(f'{location}\r\n', 201)
+    response.headers['Location'] = location
     if EXTERNAL_URL:
         response.autocorrect_location_header = False
     response.headers['Cache-Control'] = ('no-cache, no-store, must-revalidate,'
-                                ' post-check=0, pre-check=0')
+                                         ' post-check=0, pre-check=0')
     response.headers['Pragma'] = 'no-cache'
     response.headers['Expires'] = expires.isoformat() + 'Z'
+    mimetype = mimetypes.guess_type(filename)[0]
     db.query('stor.add',
              id=file_id,
-             fname=file_name,
-             mimetype=f.mimetype,
+             fname=filename,
+             sha256sum=sha256sum,
+             mimetype=mimetype if mimetype else 'application/octet-stream',
              expires=expires,
              oneshot=oneshot,
              data=contents)
@@ -110,6 +118,7 @@ def download(file_id, file_key, file_name):
         db.query('stor.expire', id=file_id)
     response = make_response(contents)
     response.headers['Content-Type'] = f['mimetype']
+    response.headers['x-hash-sha256'] = f['sha256sum']
     if val_to_boolean(request.args.get('raw')):
         response.headers[
             'Content-Disposition'] = f'attachment;filename={file_name}'
@@ -128,12 +137,10 @@ def clean_db():
 
 dbconn = db.connect()
 
-from sqlalchemy import (MetaData, Table, Column, CHAR, VARCHAR, DateTime,
-                        LargeBinary, BOOLEAN)
-
 meta = MetaData()
 stor = Table('stor', meta, Column('id', CHAR(16), primary_key=True),
              Column('fname', VARCHAR(255), nullable=False),
+             Column('sha256sum', CHAR(64), nullable=False),
              Column('mimetype', VARCHAR(255), nullable=False),
              Column('expires', DateTime(timezone=True), nullable=False),
              Column('oneshot', BOOLEAN, nullable=False, server_default='0'),
